/*
  counterdialog.cpp

  This file provides an user interface to change the counter value of an
  spdrs60 panel.

  Copyright (C) 2016 Guido Scholz <guido-scholz@gmx.net>

  This file is part of spdrs60.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or (at
  your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ams.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <QLabel>
#include <QLayout>
#include <QDialogButtonBox>
#include <QIntValidator>

#include "counterdialog.h"


CounterDialog::CounterDialog(QWidget* parent): QDialog(parent)
{
    setObjectName("CounterDialog");
    setWindowTitle(tr("Edit counter value"));

    QIntValidator* valueValidator = new QIntValidator(0, 9999, this);

    /*Layout to separate OK Cancel buttons from the upper rest*/
    QBoxLayout* baseLayout = new QVBoxLayout(this);
    baseLayout->setSpacing(10);

    /*line with text edit line*/
    QHBoxLayout* textLayout = new QHBoxLayout();
    baseLayout->addLayout(textLayout);

    QLabel* labelText = new QLabel(tr("&Counter value"));
    valueedit = new QLineEdit();
    valueedit->setMaxLength(4);
    valueedit->setValidator(valueValidator);
    labelText->setBuddy(valueedit);

    textLayout->addWidget(labelText);
    textLayout->addStretch();
    textLayout->addWidget(valueedit);

    QDialogButtonBox* buttonBox = new QDialogButtonBox(
            QDialogButtonBox::Ok |
            QDialogButtonBox::Cancel);

    connect(buttonBox, SIGNAL(accepted()), this, SLOT(accept()));
    connect(buttonBox, SIGNAL(rejected()), this,
            SLOT(reject()));
    baseLayout->addWidget(buttonBox);
}


unsigned int CounterDialog::getCounterValue()
{
    return valueedit->text().toUInt();
}


void CounterDialog::setCounterValue(unsigned int number)
{
    valueedit->setText(QString::number(number));
}

