/***************************************************************************
                           gbsarea.cpp
                           -------------------------------
    copyright            : (C) 1999-2003 by Stefan Preis
                         : (C) 2004-2009 by Guido Scholz
    email                : guido.scholz@bayernline.de
    last modified        : $Date: 2009-10-26 21:59:44 $
                           $Revision: 1.121 $
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
   this file handles loading, saving of layout files, provides an empty
   layout, sets up the elements of your layout and handles the routing
   dependant actions
 ***************************************************************************/

#include <cstdlib>            //for free, calloc, realloc

#include <QDrag>
#include <QMimeData>

#include "externalbuttonpanel.h"
#include "finder.h"
#include "panelfactory.h"
#include "resources.h"
#include "gbsarea.h"

/*cursor pixmaps*/
#include "pixmaps/cursor_wgt_b.xpm"
#include "pixmaps/cursor_wgt_m.xpm"
#include "pixmaps/cursor_wht_b.xpm"
#include "pixmaps/cursor_wht_m.xpm"
#include "pixmaps/cursor_fht_b.xpm"
#include "pixmaps/cursor_fht_m.xpm"
#include "pixmaps/cursor_ufgt_b.xpm"
#include "pixmaps/cursor_ufgt_m.xpm"
#include "pixmaps/cursor_mgt_b.xpm"
#include "pixmaps/cursor_mgt_m.xpm"
#include "pixmaps/cursor_sgt_b.xpm"
#include "pixmaps/cursor_sgt_m.xpm"
#include "pixmaps/cursor_hagt_b.xpm"
#include "pixmaps/cursor_hagt_m.xpm"

#include "pixmaps/cursor_rzs_b.xpm"
#include "pixmaps/cursor_rzs_m.xpm"
#include "pixmaps/cursor_uzs_b.xpm"
#include "pixmaps/cursor_uzs_m.xpm"
#include "pixmaps/cursor_zhs_b.xpm"
#include "pixmaps/cursor_zhs_m.xpm"
#include "pixmaps/cursor_rrs_b.xpm"
#include "pixmaps/cursor_rrs_m.xpm"
#include "pixmaps/cursor_urs_b.xpm"
#include "pixmaps/cursor_urs_m.xpm"
#include "pixmaps/cursor_paint_b.xpm"
#include "pixmaps/cursor_paint_m.xpm"
#include "pixmaps/cursor_erase_b.xpm"
#include "pixmaps/cursor_erase_m.xpm"



GBSArea::GBSArea(QWidget* parent): QWidget(parent),
    cols(0),
    rows(0),
    layoutid(0),
    layoutname(tr("Untitled")),
    tablelight(true),
    modified(false),
    gkbState(kNoneClicked),
    visualMode(kvmNormal),
    lyeditMode(lemSelect),
    paintItem(SpdrPanel::siciSt1),
    lastelement(NULL),
    SRCP08GA1InitWalker(0),
    SRCP08GA2InitWalker(0),
    SRCP08BusCount(0),
    SRCP08BusWalker(0),
    pSRCP08BusList(NULL),
    dragging(false),
    erasing(false),
    painting(false)
{
    setAcceptDrops(true);

    QPalette p(palette());
    p.setColor(backgroundRole(), Qt::lightGray);
    setPalette(p);

    /*cursor setup */
    QPixmap cb = QPixmap(cursor_wgt_b_xpm);
    QPixmap cm = QPixmap(cursor_wgt_m_xpm);
    cb.setMask(cm.mask());
    WGTCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_wht_b_xpm);
    cm = QPixmap(cursor_wht_m_xpm);
    cb.setMask(cm.mask());
    WHTCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_fht_b_xpm);
    cm = QPixmap(cursor_fht_m_xpm);
    cb.setMask(cm.mask());
    FHTCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_ufgt_b_xpm);
    cm = QPixmap(cursor_ufgt_m_xpm);
    cb.setMask(cm.mask());
    UfGTCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_mgt_b_xpm);
    cm = QPixmap(cursor_mgt_m_xpm);
    cb.setMask(cm.mask());
    MGTCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_sgt_b_xpm);
    cm = QPixmap(cursor_sgt_m_xpm);
    cb.setMask(cm.mask());
    SGTCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_hagt_b_xpm);
    cm = QPixmap(cursor_hagt_m_xpm);
    cb.setMask(cm.mask());
    HaGTCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_rzs_b_xpm);
    cm = QPixmap(cursor_rzs_m_xpm);
    cb.setMask(cm.mask());
    RZSCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_uzs_b_xpm);
    cm = QPixmap(cursor_uzs_m_xpm);
    cb.setMask(cm.mask());
    UZSCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_zhs_b_xpm);
    cm = QPixmap(cursor_zhs_m_xpm);
    cb.setMask(cm.mask());
    ZHSCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_rrs_b_xpm);
    cm = QPixmap(cursor_rrs_m_xpm);
    cb.setMask(cm.mask());
    RRSCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_urs_b_xpm);
    cm = QPixmap(cursor_urs_m_xpm);
    URSCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_paint_b_xpm);
    cm = QPixmap(cursor_paint_m_xpm);
    cb.setMask(cm.mask());
    paintCursor = QCursor(cb, 0, 0);

    cb = QPixmap(cursor_erase_b_xpm);
    cm = QPixmap(cursor_erase_m_xpm);
    cb.setMask(cm.mask());
    eraseCursor = QCursor(cb, 0, 0);

    delayTimer = new QTimer(this);
    connect(delayTimer, SIGNAL(timeout()),
            this, SLOT(slotElementClickedTimeout()));
}


GBSArea::~GBSArea()
{
    /* delete all elements*/
    for (int i = 0; i < elements.size(); ++i) {
        delete elements[i];
        elements[i] = NULL;
    }
    elements.clear();

    /* clear SRCP bus lists */
    if ((SRCP08BusCount > 0) && (pSRCP08BusList != NULL))
        free(pSRCP08BusList);

}


// *INDENT-OFF*
QSize GBSArea::sizeHint() const
{
    return QSize(1 + cols * (EL_WIDTH + 1), 1 + rows * (EL_HEIGHT + 1));
}
// *INDENT-ON*


void GBSArea::newFile(int newcols, int newrows, unsigned int newid,
        const QString& newname)
{
    // delete all old routes and layout data
    emit clearRoutes();
    emit updateRoutingViewer("");

    /* delete all elements*/
    for (int i = 0; i < elements.size(); ++i) {
        delete elements[i];
        elements[i] = NULL;
    }
    elements.clear();

    cols = newcols;
    rows = newrows;
    layoutid = newid;
    layoutname = newname;
    tablelight = true;

    elements.resize(rows * cols);
    adjustSize();
    setModified(false);
}


void GBSArea::writeFileTextToStream(QTextStream& ts)
{
    ts << "# start of layout section" << Qt::endl
       << "%% layout" << Qt::endl
       << "# layout dimensions=columns" << DS << "rows" << Qt::endl
       << GF_DIMENSIONS << DS << cols << DS << rows << Qt::endl
       << GF_ID << DS << layoutid << DS << layoutname << Qt::endl
       << GF_TABLELIGHT << DS << (tablelight ? 1 : 0) << Qt::endl
       << "# start of element section" << Qt::endl;
       //<< "# elements=" << elements.count() << Qt::endl;

    for (int i = 0; i < elements.size(); i++) {
        if (elements[i] != NULL) {
            ts << "%% element " << i << Qt::endl;
            elements[i]->writeFileTextToStream(ts);
        }
    }
    setModified(false);
}


void GBSArea::readFileTextFromStream(QTextStream& ts)
{
    PanelFactory pf;

    /*clear old element list*/
    if (!elements.isEmpty()) {
        emit clearRoutes();

        /* delete all elements*/
        for (int i = 0; i < elements.size(); ++i) {
            delete elements[i];
            elements[i] = NULL;
        }
        elements.clear();
    }
    /*set default value to make light switchable*/
    tablelight = true;

    QString s, key, value;
    unsigned int ecount = 0;
    bool tablelighton = true;

    while (!ts.atEnd()) {
        s = ts.readLine();

        /* ignore comment lines */
        if (s.startsWith("#"))
            continue;

        /* read layout dimensions */
        /* key/value pairs are read sequence independent */
        key = s.section(DS, 0, 0);
        value = s.section(DS, 1, 1).trimmed();

        if (key.compare(GF_DIMENSIONS) == 0) {
            cols = value.toInt();
            value = s.section(DS, 2, 2).trimmed();
            rows = value.toInt();
            ecount = cols * rows;
            elements.resize(ecount);
            adjustSize();
        }
        else if (key.compare(GF_ID) == 0) {
            layoutid = value.toUInt();
            layoutname = s.section(DS, 2).trimmed();
        }
        else if (key.compare(GF_TABLELIGHT) == 0) {
            tablelighton = (value.toInt() == 1);
        }

        /*here we read allways up to start marker of a new route*/
        else if (s.startsWith("%% element")) {

            // read next line containing classid
            s = ts.readLine();
            SpdrPanel::SpdrItemClassId type =
                (SpdrPanel::SpdrItemClassId) s.section(';', 1, 1).toInt();

            SpdrPanel* el = pf.createPanelFromStream(ts, this, type);
            if (el != NULL) {
                unsigned int idx = el->getIndexNo();

                if (idx < ecount) {
                    moveElementToIndexPos(el, idx);
                    elements.insert(idx, el);
                    connectElement(el);
                    el->show();
                }
                else {
                    qWarning("Error: Element outside of layout found "
                            "(Index = %d).", idx);
                    delete el;
                }
                //qWarning("Element number %d inserted.", el->getIndexNo());
            }
        }

        /*here we read allways up to start marker of a new route*/
        else if (s.startsWith("%% panel")) {

            QStringList tokens = s.split(' ');
            SpdrPanel::SpdrItemClassId type =
                (SpdrPanel::SpdrItemClassId) tokens[2].toInt();

            SpdrPanel* el = pf.createPanelFromStream(ts, this, type);
            if (el != NULL) {
                unsigned int idx = el->getIndexNo();

                if (idx < ecount) {
                    moveElementToIndexPos(el, idx);
                    elements[idx] = el;
                    connectElement(el);
                    el->show();
                }
                else {
                    qWarning("Error: Element outside of layout found "
                            "(Index = %d).", idx);
                    delete el;
                }
                //qWarning("Element number %d inserted.", el->getIndexNo());
            }
        }
        else if (s.startsWith("%% route"))
            break;
        //TODO:
        /*else if (s.startsWith("%% fbsection"))
            break;*/
    }

    /*switch table light only off if this state was saved, default is on*/
    if (!tablelighton)
        switchTableLight(false);

    setModified(false);
}


void GBSArea::slotFHTclicked()
{
    externalButtonClicked(kFhtClicked);
}


void GBSArea::slotFRTclicked()
{
    externalButtonClicked(kFrtClicked);
}


void GBSArea::slotWGTclicked()
{
    externalButtonClicked(kWgtClicked);
}


void GBSArea::slotUfGTclicked()
{
    externalButtonClicked(kUfgtClicked);
}


void GBSArea::slotMGTclicked()
{
    externalButtonClicked(kMgtClicked);
}


void GBSArea::slotSGTclicked()
{
    externalButtonClicked(kSgtClicked);
}


void GBSArea::slotHaGTclicked()
{
    externalButtonClicked(kHagtClicked);
}


void GBSArea::externalButtonClicked(GbsButtonState externalButton)
{
    /*
     * store pressed control button, activate cursor and start timer
     */

    switch (externalButton) {
        case kFhtClicked:
            delayTimer->start(cDelayTime);
            gkbState = externalButton;
            setCursor(FHTCursor);
            break;
        case kMgtClicked:
            delayTimer->start(cDelayTime);
            gkbState = externalButton;
            setCursor(MGTCursor);
            /*TODO: implement MGT-function*/
            QApplication::beep();
            emit statusMessage(tr("MGT-Function not supported."));
            delayTimer->start(500);
            break;
        case kUfgtClicked:
            delayTimer->start(cDelayTime);
            gkbState = externalButton;
            setCursor(UfGTCursor);
            break;
        case kWgtClicked:
            delayTimer->start(cDelayTime);
            gkbState = externalButton;
            setCursor(WGTCursor);
            break;
        case kWhtClicked:
            delayTimer->start(cDelayTime);
            gkbState = externalButton;
            setCursor(WHTCursor);
            break;
        case kSgtClicked:
            delayTimer->start(cDelayTime);
            gkbState = externalButton;
            setCursor(SGTCursor);
            break;
        case kHagtClicked:
            delayTimer->start(cDelayTime);
            gkbState = externalButton;
            setCursor(HaGTCursor);
            break;
        case kEinClicked:
            if (delayTimer->isActive())
                delayTimer->stop();
            slotElementClickedTimeout();
            switchTableLight(true);
            break;
        case kAusClicked:
            if (delayTimer->isActive())
                delayTimer->stop();
            slotElementClickedTimeout();
            switchTableLight(false);
            break;

        default:
            break;
    }
}


void GBSArea::slotElementClicked(element* el, GbsButtonState gbsButton)
{
    /* stop running timer but do not change cursor shape until we know
     * what exactly happens next */

    if (delayTimer->isActive())
        delayTimer->stop();

    switch (gbsButton) {

        /*signal buttons*/
        case kRfsClicked:
        case kZfsClicked:
        case kZhsClicked:
            if (kSgtClicked == gkbState || kHagtClicked == gkbState) {
                if (el->isLocked()) {
                    QApplication::beep();
                    emit statusMessage(tr("No switching possible, "
                                "signal '%1' is locked by an active route.")
                            .arg(el->getLabelText()));
                }
                else
                    if (kSgtClicked == gkbState)
                        el->switchToDir(1);
                    else
                        el->switchToDir(0);
                slotElementClickedTimeout();
            }
            else if (kFhtClicked == gkbState) {
                // give direct button pressed feedback to user
                setCursor(Qt::ArrowCursor);
                /* send signal to route controller*/
                emit resetRoute(el, gbsButton);
            }
            else if (kUfgtClicked == gkbState || kNoneClicked ==
                    gkbState) {
                /*if (el->hasFfMLock()) {
                    QApplication::beep();
                    emit statusMessage(tr("No routing possible; signal '%1'"
                                " is already locked by an active route.")
                            .arg(el->getLabelText()));
                    slotElementClickedTimeout();
                }
                else {*/
                    // give button pressed feedback to user
                    setCursor(Qt::ArrowCursor);
                    /* send signal to route controller*/
                    emit setRoute(el, gbsButton, gkbState);
                //}
            }
            else if (kWgtClicked == gkbState) {
                QApplication::beep();
                emit statusMessage(tr("Signals can not be switched "
                            "using WGT"));
                slotElementClickedTimeout();
            }
            else {
                QApplication::beep();
                emit statusMessage(tr("Operation not allowed"));
                slotElementClickedTimeout();
            }
            break;

        /*turnout button*/
        case kTurnoutClicked:
            if (kWgtClicked == gkbState) {
                if (el->isLocked()) {
                    QApplication::beep();
                    emit statusMessage(tr("No switching possible, "
                                "solenoid '%1' is locked by an active route.")
                            .arg(el->getLabelText()));
                }
                else if (el->isOccupied()) {
                    QApplication::beep();
                    emit statusMessage(tr("No switching possible, "
                                "turnout is occupied"));
                }
                else
                    el->toggle();
            }

            else if (kWhtClicked == gkbState) {
                if (el->isLocked()) {
                    QApplication::beep();
                    emit statusMessage(tr("No switching possible, "
                                "solenoid '%1' is locked by an active route.")
                            .arg(el->getLabelText()));
                }
                else
                    el->toggle();
            }

            else if (kFhtClicked == gkbState) {
                QApplication::beep();
                emit statusMessage(tr("Turnouts can not be switched "
                            "using FHT"));
            }
            else if (kSgtClicked == gkbState) {
                QApplication::beep();
                emit statusMessage(tr("Turnouts can not be switched "
                            "using SGT"));
            }
            else if (kUfgtClicked == gkbState) {
                QApplication::beep();
                emit statusMessage(tr("Turnouts can not be switched "
                            "using UfGT"));
            }
            else {
                QApplication::beep();
                emit statusMessage(tr("Operation not allowed"));
            }
            slotElementClickedTimeout();
            break;

        default:
            break;
    }
}


void GBSArea::startRouteTimer(Route::RouteType tor)
{
    switch (tor) {
        case Route::rtRZS:
            setCursor(RZSCursor);
            break;
        case Route::rtUZS:
            setCursor(UZSCursor);
            break;
        case Route::rtZHS:
            setCursor(ZHSCursor);
            break;
        case Route::rtRRS:
            setCursor(RRSCursor);
            break;
        case Route::rtURS:
            setCursor(URSCursor);
            break;
    }
    delayTimer->start(cDelayTime);
}


void GBSArea::slotElementClickedTimeout()
{
    emit resetSelectedSignal();
    gkbState = kNoneClicked;
    setCursor(Qt::ArrowCursor);
}


void GBSArea::updateRoutePathLEDs(const stateElement& fSig,
        const stateElement& tSig, Route::RouteSetAction& setRoute)
{
    if (fSig.elemPtr == NULL || tSig.elemPtr == NULL)
        return;

    if (rows <= 1)
        return;

    int idx = fSig.elemPtr->getIndexNo();
    int maxIdx = (int) elements.size();
    bool finished = false;
    bool setrt = (Route::rsaReset != setRoute);

    unsigned int entrydir;
    unsigned int exitdir;

    entrydir = fSig.elemPtr->entryDir();

    while (!finished) {
        if ((idx < 0) || (idx >= maxIdx)) {
            emit statusMessage(tr("Route hit layout edge at element %1")
                    .arg(idx));
            break;
        }

        element* rel = dynamic_cast<element*>(elements[idx]);
        if (rel == NULL) {
            emit statusMessage(tr("Route end at empty element %1")
                    .arg(idx));
            break;
        }

        if (!rel->isRoutable()) {
            emit statusMessage(tr("Route end at not routable element %1")
                    .arg(idx));
            break;
        }

        // interrupt operation when normal route meets occupied element
        // start signal is allowed to be occupied
        // give also feedback for caller
        if ((Route::rsaZfs == setRoute) && rel->isOccupied() &&
                rel != fSig.elemPtr) {
            setRoute = Route::rsaReset;
            return;
        }

        // paint or remove yellow track and get back new route direction
        // rel->setRouted(setrt);
        // rel->getExitDirection(entrydir);
        exitdir = rel->routeElement(entrydir, setrt);

        if (exitdir == rdCenter) {
            emit statusMessage(tr("Route found dead end at element %1")
                    .arg(idx));
            break;
        }

        entrydir = rdCenter;

        if (exitdir & rdN) {
            --idx;
            entrydir = entrydir | rdS;
        }
        else if (exitdir & rdS) {
            ++idx;
            entrydir = entrydir | rdN;
        }
        if (exitdir & rdW) {
            idx -= rows;
            entrydir = entrydir | rdE;
        }
        else if (exitdir & rdE) {
            idx += rows;
            entrydir = entrydir | rdW;
        }

        finished = (rel == tSig.elemPtr);
    }
}


void GBSArea::slotEditFind(const QString& ftext, int type, int multiple)
{
    bool found = findElement(ftext, type, multiple);

    // no element could be located -> show this information
    if (!found)
        QMessageBox::information(this, tr("Locate error"),
                                 tr("There's no element which\n"
                                    "matches your search criteria."));
}


/**
 * Search all elements for a matching address or text label
 */
bool GBSArea::findElement(const QString& ftext, int ftype, int fmulti)
{
    QString s = "";
    int idx;
    bool returnvalue = false;

    // locate element with certain address 1
    if (ftype == SRCH_A1)
        for (idx = 0; idx < elements.size(); idx++) {
            element* el = dynamic_cast<element*>(elements[idx]);

            if (el == NULL)
                continue;

            if (ftext.toInt() == el->getAddress1()) {
                returnvalue = true;
                el->locateMe();
                if (fmulti == SRCH_MULTI)
                    continue;
                else
                    return returnvalue;
            }
        }

    // locate element with certain address 2
    else if (ftype == SRCH_A2)
        for (idx = 0; idx < elements.size(); idx++) {
            element* el = dynamic_cast<element*>(elements[idx]);

            if (el == NULL)
                continue;

            if (ftext.toInt() == el->getAddress2()) {
                returnvalue = true;
                el->locateMe();
                if (fmulti == SRCH_MULTI)
                    continue;
                else
                    return returnvalue;
            }
        }

    // locate element with certain textfield
    else if (ftype == SRCH_TX) {
        for (idx = 0; idx < elements.size(); idx++) {
            element* el = dynamic_cast<element*>(elements[idx]);

            if (el == NULL)
                continue;

            s = el->sSoldText;
            if (s.contains(ftext)) {
                returnvalue = true;
                el->locateMe();
                if (fmulti == SRCH_MULTI)
                    continue;
                else
                    return returnvalue;
            }
        }
    }
    return returnvalue;
}


/**
 * Toggle all elements but no couplers, motors no shifting bridges,
 * no turntables
 **/
void GBSArea::slotToggleAll()
{
    for (int i = 0; i < elements.size(); i++) {
        element* el = dynamic_cast<element*>(elements[i]);

        if (el != NULL &&
                el->classId() != SpdrPanel::siciEnk &&
                el->classId() != SpdrPanel::siciMdc &&
                el->classId() != SpdrPanel::siciSbn &&
                el->classId() != SpdrPanel::siciDre)
            el->toggle();
    }
}


/**
 *  Send current state off all element to SRCP server
 **/
void GBSArea::slotSendAll()
{
    for (int i = 0; i < elements.size(); i++) {
        element* el = dynamic_cast<element*>(elements[i]);

        if (el != NULL &&
                el->classId() != SpdrPanel::siciEnk &&
                el->classId() != SpdrPanel::siciMdc &&
                el->classId() != SpdrPanel::siciSbn &&
                el->classId() != SpdrPanel::siciDre)
            el->sendSrcpState();
    }
}


/**
 * set all signals to red state
 * */
void GBSArea::slotNotrot()
{
    for (int i = 0; i < elements.size(); i++) {
        element* el = dynamic_cast<element*>(elements[i]);

        if (el != NULL && el->isSignal())
            el->switchToDir(0);
    }

    emit statusMessage(tr("Switched all signals to halt/stop"));
}


/**
 * connect element to gbs events
 *
 * disconnect is done automatically by the Qt library when element
 * is deleted
 */
void GBSArea::connectElement(SpdrPanel* sp)
{
    if (sp == NULL)
        return;

    /*all spdrpanels*/
    connect(this, SIGNAL(switchedVisualMode(elemVisualMode)),
            sp, SLOT(switchVisualMode(elemVisualMode)));
    /*tooltips on/off, address labeling on/off*/
    connect(this, SIGNAL(sigRepaintLayout()),
            sp, SLOT(slotRepaintLayout()));

    /*connect external button panels*/
    ExternalButtonPanel* ebp = dynamic_cast<ExternalButtonPanel*>(sp);
    if (ebp != NULL) {

        /*outgoing button clicks*/
        connect(ebp, SIGNAL(buttonClicked(GbsButtonState)),
                this, SLOT(externalButtonClicked(GbsButtonState)));

        /*outgoing feedback trigger dialog with main window*/
        connect(ebp, SIGNAL(sigShowFBmodules()),
                this, SIGNAL(sigShowFBmodules()));

        /*incomming feedback changes*/
        connect(this, SIGNAL(feedbackPortChanged(unsigned int,
                        unsigned int, bool)),
                ebp, SLOT(slotOccupyElement(unsigned int,
                        unsigned int, bool)));
    }

    /*connect undifferentiated items*/
    element* el = dynamic_cast<element*>(sp);
    if (el != NULL) {
        /*button panels*/
        connect(el, SIGNAL(elementClicked(element*, GbsButtonState)),
                this, SLOT(slotElementClicked(element*, GbsButtonState)));

        /*for route members*/
        connect(el, SIGNAL(recordElement(element*, elemRecordType)),
                this, SIGNAL(recordElement(element*, elemRecordType)));

        /*feedback handling panels, connect feedback trigger dialog with
          main window*/
        connect(el, SIGNAL(sigShowFBmodules()),
                this, SIGNAL(sigShowFBmodules()));

        /*SRCP message handling for panels with addresses or feedback
          contacts */
        connect(this, SIGNAL(sigShowElement(int, int,
                        elemSelectionMode)),
                sp, SLOT(slotShowElement(int, int, elemSelectionMode)));
        connect(this, SIGNAL(processInfoPortMessage(unsigned int,
                        unsigned int, unsigned int, unsigned int)),
                el, SLOT(processInfoPortMessage(unsigned int,
                        unsigned int, unsigned int, unsigned int)));
        connect(el, SIGNAL(sendSrcpMessage(SrcpMessage*)),
                this, SIGNAL(sendSrcpMessage(SrcpMessage*)));
        connect(this, SIGNAL(feedbackPortChanged(unsigned int,
                        unsigned int, bool)),
                el, SLOT(slotOccupyElement(unsigned int,
                        unsigned int, bool)));
        connect(this, SIGNAL(feedbackPortChanged(unsigned int,
                        unsigned int, bool, unsigned int)),
                el, SLOT(slotOccupyElement(unsigned int,
                        unsigned int, bool, unsigned int)));
    }
}


// *INDENT-OFF*
bool GBSArea::isModified()
// *INDENT-ON*
{
    for (int i = 0; i < elements.size(); i++) {
        SpdrPanel* el = elements[i];
        if (el != NULL) {
            if (el->isModified()) {
                modified = true;
                break;
            }
        }
    }
    return modified;
}


void GBSArea::setModified(bool m)
{
    if (modified != m)
        modified = m;

    //TODO: check this, should be updated if element was
    //  - removed
    //  - added
    //  - edited
    updateSRCP08BusList();
}


/*return number of layout columns*/
int GBSArea::getColumns()
{
    return cols;
}


/*return number of layout rows*/
int GBSArea::getRows()
{
    return rows;
}


/*set new layout size*/
void GBSArea::setLayoutSize(int newcols, int newrows)
{
    if (newcols < 1)
        newcols = 1;

    if (newrows < 1)
        newrows = 1;

    if (newcols > MAX_COLS)
        newcols = MAX_COLS;

    if (newrows > MAX_ROWS)
        newrows = MAX_ROWS;

    QVector<SpdrPanel*> tmpelements;

    tmpelements.resize(newcols * newrows);

    /* delete elements outside new gbs area */
    if (newcols < cols) {
        for (int i = 0; i < (cols - newcols); i++) {
            removeColumnElements(cols - i);
        }
    }

    if (newrows < rows) {
        for (int i = 0; i < (rows - newrows); i++) {
            removeRowElements(rows - i);
        }
    }

    /* rearrange elements in altered gbs */
    for (int c = 1; c <= newcols; c++) {
        for (int r = 1; r <= newrows; r++) {
            SpdrPanel* el = item(r, c);

            if (el != NULL) {
                unsigned int idx = newrows * (c - 1) + r - 1;
                el->setIndexNo(idx);
                tmpelements[idx] = el;
            }
        }
    }

    elements = tmpelements;

    cols = newcols;
    rows = newrows;

    adjustSize();
    setModified(true);
}


/*set new layout id*/
void GBSArea::setLayoutId(unsigned int newid)
{
    if (layoutid != newid) {
        layoutid = newid;
        modified = true;
    }
}


/*return layout id*/
unsigned int GBSArea::getLayoutId()
{
    return layoutid;
}


/*set new layout name*/
void GBSArea::setLayoutName(const QString& newname)
{
    if (layoutname != newname) {
        layoutname = newname;
        modified = true;
    }
}


/*return layout name*/
QString GBSArea::getLayoutName() const
{
    return layoutname;
}


/*remove all elements of a single column*/
void GBSArea::removeColumnElements(int col)
{
    for (int r = 1; r <= rows; r++) {
        int idx = indexOf(r, col);
        delete elements[idx];
        elements[idx] = NULL;
    }
}

/*remove all elements of a single row*/
void GBSArea::removeRowElements(int row)
{
    for (int c = 1; c <= cols; c++) {
        int idx = indexOf(row, c);
        delete elements[idx];
        elements[idx] = NULL;
    }
}

/*
 *  row is defined from 1 to rows
 *  col is defined from 1 to cols
 */
// *INDENT-OFF*
int GBSArea::indexOf(int row, int col) const
{
    return (rows * (col - 1) + row - 1);
}
// *INDENT-ON*


// *INDENT-OFF*
int GBSArea::indexOf(const QPoint& ep) const
{
    int row = ep.y() / (EL_HEIGHT + 1) + 1;
    int col = ep.x() / (EL_WIDTH + 1) + 1;
    return (rows * (col - 1) + row - 1);
}
// *INDENT-ON*


void GBSArea::moveElementToIndexPos(SpdrPanel* el, int idx)
{
    if (el != NULL && idx < elements.size())
        el->move(1 + (idx / rows) * (EL_WIDTH + 1),
                1 + (idx % rows) * (EL_HEIGHT + 1));
}


// *INDENT-OFF*
SpdrPanel* GBSArea::item(int row, int col) const
// *INDENT-ON*
{
    if (row < 0 || col < 0 || row > rows ||
            col > cols || row * col >= elements.size())
        return NULL;

    return elements[indexOf(row, col)];
}


QVector<SpdrPanel*>* GBSArea::getGbsElementListPtr()
{
    return &elements;
}


void GBSArea::sendInfoPortMessage(unsigned int bus,
        unsigned int addr, unsigned int port, unsigned int value)
{
    /* send incomming GA actions to all elements*/
    emit processInfoPortMessage(bus, addr, port, value);
}

/*used by route dialog to look for element names*/
void GBSArea::getElementByAddress(const int bus, const int address,
        element** el)
{
    for (int i = 0; i < elements.size(); i++) {
        element* gbse = dynamic_cast<element*>(elements[i]);

        if (gbse != NULL && gbse->hasSameAddress(bus, address)) {
            *el = gbse;
            break;
        }
    }
}


bool GBSArea::runSRCP08GAInitSequence()
{
    bool returnvalue = false;
    bool CounterChanged = false;

    for (int i = SRCP08GA1InitWalker; i < elements.size(); i++) {

        element* el = dynamic_cast<element*>(elements[i]);
        if (el == NULL)
            continue;

        if (!el->isSwitchable())
            continue;

        if (el->sendSRCP08InitGA(1)) {
            SRCP08GA1InitWalker = i;
            SRCP08GA1InitWalker++;
            returnvalue = (SRCP08GA1InitWalker < elements.size());
            CounterChanged = true;
            break;
        }
    }

    if (!returnvalue && !CounterChanged)
        for (int i = SRCP08GA2InitWalker; i < elements.size(); i++) {

            element* el = dynamic_cast<element*>(elements[i]);
            if (el == NULL)
                continue;

            if (!el->isSwitchable() || el->getAddressCount() != 2)
                continue;

            if (el->sendSRCP08InitGA(2)) {
                SRCP08GA2InitWalker = i;
                SRCP08GA2InitWalker++;
                returnvalue = (SRCP08GA2InitWalker < elements.size());
                break;
            }
        }

    /* reset when init process for all GAs is finished */
    if (!returnvalue) {
        SRCP08GA1InitWalker = 0;
        SRCP08GA2InitWalker = 0;
    }

    return returnvalue;
}


/*
 * ask server about power status off every single bus, but only one at
 * a time return true while there are unasked busses left
 */
bool GBSArea::sendSRCP08BusMessage(SrcpMessage::Type smt)
{
    bool WalkerChanged = false;

    /*
     * all busses are asked step by step, allways waiting
     * for server OK
     */
    if ((SRCP08BusCount > 0)
        && (SRCP08BusWalker < SRCP08BusCount)) {

        //SrcpMessage* sm = new SrcpMessage(SrcpMessage::msgPowerGet);
        PowerMessage sm = PowerMessage(smt,
                pSRCP08BusList[SRCP08BusWalker].id);
        emit sendSrcpMessage(&sm);

        SRCP08BusWalker++;
        WalkerChanged = true;
    }

    /* reset when this process is finished for all busses */
    if (!WalkerChanged)
        SRCP08BusWalker = 0;

    return WalkerChanged;
}


/*
 * switch on every single bus, but only one at a time
 * return true while there are unswitched busses left
 */
bool GBSArea::setSRCP08BusPower(bool setPowerOn)
{
    bool CounterChanged = false;

    /*
     * all busses are switched step by step, allways waiting
     * for server OK
     */
    if ((SRCP08BusCount > 0)
        && (SRCP08BusWalker < SRCP08BusCount)) {
        PowerMessage sm = PowerMessage(SrcpMessage::mtPowerSet,
                pSRCP08BusList[SRCP08BusWalker].id, setPowerOn);
        emit sendSrcpMessage(&sm);

        SRCP08BusWalker++;
        CounterChanged = true;
    }

    /* reset when init process for all busses is finished */
    if (!CounterChanged)
        SRCP08BusWalker = 0;

    return CounterChanged;
}


void GBSArea::updateSRCP08BusList()
{
    unsigned int count = 0, busno = 0;
    bool busNoIsKnown;
    SrcpBus *tempbuslist;

    /*how many different GA busses do we have? */
    for (int i = 0; i < elements.size(); i++) {

        element* el = dynamic_cast<element*>(elements[i]);
        if (el == NULL)
            continue;

        for (int k = 0; k <= 2; k++) {
            switch (k) {
                case 0:
                    busno = el->getGA1BusNo();
                    break;
                case 1:
                    busno = el->getGA2BusNo();
                    break;
                case 2:
                    busno = el->getFBBusNo();
                    break;
            }

            if (busno > 0) {
                if (count == 0) {
                    count++;

                    pSRCP08BusList = (SrcpBus *) calloc(count,
                            sizeof(SrcpBus));
                    if (pSRCP08BusList == NULL) {
                        qWarning("Memory allocation error!");
                        SRCP08BusCount = count - 1;
                        return;
                    }
                    pSRCP08BusList[0].id = busno;
                    pSRCP08BusList[0].hasFb = false;
                    pSRCP08BusList[0].hasGa = false;

                    if (k == 2)
                        pSRCP08BusList[0].hasFb = true;
                    else
                        pSRCP08BusList[0].hasGa = true;
                }
                /*count > 0 */
                else {
                    busNoIsKnown = false;
                    for (unsigned int j = 0; j < count; j++) {
                        if (pSRCP08BusList[j].id == busno) {
                            busNoIsKnown = true;
                            if (k == 2)
                                pSRCP08BusList[count - 1].hasFb = true;
                            else
                                pSRCP08BusList[count - 1].hasGa = true;
                            break;
                        }
                    }
                    if (!busNoIsKnown) {
                        count++;

                        tempbuslist =
                            (SrcpBus *) realloc(pSRCP08BusList,
                                            sizeof(SrcpBus) * count);

                        if (tempbuslist == NULL) {
                            qWarning("Memory allocation error!");
                            SRCP08BusCount = count - 1;
                            return;
                        }
                        pSRCP08BusList = tempbuslist;
                        pSRCP08BusList[count - 1].id = busno;
                        pSRCP08BusList[count - 1].hasFb = false;
                        pSRCP08BusList[count - 1].hasGa = false;

                        if (k == 2)
                            pSRCP08BusList[count - 1].hasFb = true;
                        else
                            pSRCP08BusList[count - 1].hasGa = true;
                    }
                }
            }   /* for busno */
        }       /* for k */
    }           /* for i */

    if (count == 0)
        statusMessage(tr("Layout does not contain a SRCP bus"
                    " configuration"));
    else if (count == 1)
        statusMessage(tr("Layout contains 1 configured bus"));
    else
        statusMessage(tr("Layout contains %1 configured busses").
                   arg(count));

    SRCP08BusCount = count;
}


bool GBSArea::hasSrcp08Bus(unsigned int bus)
{
    bool returnvalue = false;

    if (SRCP08BusCount > 0)
        for (unsigned int i = 0; i < SRCP08BusCount; i++)
            if (bus == pSRCP08BusList[i].id) {
                returnvalue = true;
                break;
            }

    return returnvalue;
}


void GBSArea::mouseReleaseEvent(QMouseEvent* e)
{
    /*normal mode and route edit mode*/
    if (visualMode == kvmNormal || visualMode == kvmEditRoute) {
            e->ignore();
    }

    /*layout edit mode*/
    else if (visualMode == kvmEditLayout) {
        if (e->button() == Qt::LeftButton) {

            if (lyeditMode == lemSelect) {
                /*TODO: handle drop action*/
                e->ignore();
            }

            else if (lyeditMode == lemErase) {
                erasing = false;
            }

            else if (lyeditMode == lemPaint) {
                painting = false;
            }

            e->accept();
        }
        else
            e->ignore();
    }
}

/**
 * start dragging of an element
 * */
void GBSArea::mousePressEvent(QMouseEvent* e)
{
    PanelFactory pf;

    if (visualMode == kvmNormal) {
        e->ignore();
    }

    /*layout edit mode*/
    else if (visualMode == kvmEditLayout) {
        if (e->button() == Qt::LeftButton) {

            /* drag item */
            if (lyeditMode == lemSelect) {
                if (qobject_cast<SpdrPanel*>(childAt(e->pos())) != NULL)
                    dragging = true;
            }

            /* erase item */
            else if (lyeditMode == lemErase) {
                if (qobject_cast<SpdrPanel*>(childAt(e->pos())) != NULL) {
                    int idx = indexOf(e->pos());
                    delete elements[idx];
                    elements[idx] = NULL;
                    modified = true;
                }
                erasing = true;
            }

            /* paint item */
            else if (lyeditMode == lemPaint) {
                SpdrPanel* el = qobject_cast<SpdrPanel*>(childAt(e->pos()));
                int idx = indexOf(e->pos());

                if (el != NULL) {
                    if (el->classId() != paintItem) {
                        delete elements[idx];
                        elements[idx] = NULL;
                        el = pf.createPanel(this, paintItem);
                        el->setIndexNo(idx);
                        moveElementToIndexPos(el, idx);
                        elements[idx] = el;
                        connectElement(el);
                        el->show();
                        modified = true;
                    }
                }
                else {
                    el = pf.createPanel(this, paintItem);
                    el->setIndexNo(idx);
                    moveElementToIndexPos(el, idx);
                    elements[idx] = el;
                    connectElement(el);
                    el->show();
                    modified = true;
                }
                painting = true;
            }
            e->accept();
        }
    }
}

/**
 * take element icon and its date, move it around
 */
void GBSArea::mouseMoveEvent(QMouseEvent* e)
{
    PanelFactory pf;

    /*layout edit mode*/
    if (visualMode == kvmEditLayout) {
        /* drag elements */
        if ((lyeditMode == lemSelect) && dragging) {
            SpdrPanel* el = (SpdrPanel*)childAt(e->pos());
            if (el != NULL) {
                unsigned int idx = el->getIndexNo();
                QByteArray data(sizeof(idx), '\0');
                memcpy(data.data(), &idx, sizeof(idx));

#if QT_VERSION >= 0x050000
                QPixmap pm = el->grab();
#else
                QPixmap pm = QPixmap::grabWidget(el);
#endif
                QMimeData *mimeData = new QMimeData;
                mimeData->setData(MIME_LE, data);
                mimeData->setImageData(pm.toImage());

                QDrag *drag = new QDrag(this);
                drag->setMimeData(mimeData);
                drag->setPixmap(pm);
                drag->setHotSpot(QPoint(EL_WIDTH / 2, EL_HEIGHT / 2));
                drag->exec(Qt::MoveAction);
                //Qt::DropAction dropAction = drag->exec();

                dragging = false;
            }
        }

        /* erase elements */
        else if ((lyeditMode == lemErase) && erasing) {
                SpdrPanel* el = (SpdrPanel*)childAt(e->pos());

                if (el != NULL) {
                    int idx = indexOf(e->pos());
                    delete elements[idx];
                    elements[idx] = NULL;
                    modified = true;
                }
        }

        /* paint elements */
        else if ((lyeditMode == lemPaint) && painting) {
                int idx = indexOf(e->pos());

                if (idx > (rows * cols))
                    return;

                SpdrPanel* el = (SpdrPanel*)childAt(e->pos());

                if (el != NULL) {
                    if (el->classId() != paintItem) {
                        delete elements[idx];
                        elements[idx] = NULL;
                        el = pf.createPanel(this, paintItem);
                        el->setIndexNo(idx);
                        moveElementToIndexPos(el, idx);
                        elements[idx] = el;
                        connectElement(el);
                        el->show();
                        modified = true;
                    }
                }
                else {
                    el = pf.createPanel(this, paintItem);
                    el->setIndexNo(idx);
                    moveElementToIndexPos(el, idx);
                    elements[idx] = el;
                    connectElement(el);
                    el->show();
                    modified = true;
                }
        }
        e->accept();
    }
}

/**
 * give feedback if this widget cares about the offered data
 * */
void GBSArea::dragEnterEvent(QDragEnterEvent* e)
{
    if (visualMode != kvmEditLayout)
        return;

    //TODO: enable DnD between different windows
    if (e->mimeData()->hasFormat(MIME_LE) && e->source() == this)
        e->acceptProposedAction();
}

/**
 * take the dropped data and do something usefull with it
 * */
void GBSArea::dropEvent(QDropEvent *e)
{
    if (visualMode != kvmEditLayout)
        return;

    if (e->mimeData()->hasFormat(MIME_LE)) {
        // decode data and insert element
        QByteArray data = e->mimeData()->data(MIME_LE);
        unsigned int idx = 0;

        if (data.size() != sizeof(idx))
            return;

        memcpy(&idx, data.data(), sizeof(idx));

        // move element from old position to new position
        SpdrPanel* el = elements[idx];
        elements[idx] = NULL;

        if (el != NULL) {
            int pidx = indexOf(e->pos());
            el->setIndexNo(pidx);
            moveElementToIndexPos(el, pidx);
            delete elements[pidx];
            elements[pidx] = el;
            modified = true;
        }
        e->acceptProposedAction();
    }
}


void GBSArea::switchVisualMode(elemVisualMode vm)
{
    if (vm != visualMode) {

        /*care for mouse pointer switching*/
        if (vm == kvmEditLayout)
            changeLayoutEditMode(lyeditMode);
        else if (visualMode == kvmEditLayout)
            setCursor(Qt::ArrowCursor);

        visualMode = vm;
        emit switchedVisualMode(vm);
        update();
    }
}

/* paint colored lines */
void GBSArea::paintEvent(QPaintEvent*)
{
    QPainter p(this);

    /* paint visual mode lines*/
    switch (visualMode) {
        case kvmEditLayout:
            // edit mode: red
            p.setPen(QColor(Qt::red));
            break;
        case kvmEditRoute:
            // show route mode: blue
            p.setPen(QColor(Qt::blue));
            break;
        default:
            // normal mode: grey
            p.setPen(QColor(Qt::gray));
            break;
    }

    int h = height();
    int w = width();

    for (int x = 0; x < w; x += (EL_WIDTH + 1))
        p.drawLine(x, 0, x, h - 1);

    for (int y = 0; y < h; y += (EL_HEIGHT + 1))
        p.drawLine(0, y, w - 1, y);
}

/*
 * change current edit mode
 */
void GBSArea::changeLayoutEditMode(LayoutEditMode lem)
{
    if (lem != lyeditMode)
        lyeditMode = lem;

    switch(lem) {
        case lemSelect:
            setCursor(Qt::ArrowCursor);
            break;
        case lemPaint:
            setCursor(paintCursor);
            break;
        case lemErase:
            setCursor(eraseCursor);
            break;
    }
}

/*
 * change current paint item
 */
void GBSArea::changeLayoutPaintItem(SpdrPanel::SpdrItemClassId sici)
{
    if (sici != paintItem)
        paintItem = sici;
}


/*
 * respond to incomming Generic Messages
 */
void GBSArea::processGenericMessage(unsigned int sendto,
        unsigned int replyto, const CrcfMessage* cm)
{
    QString cms("");

    if (NULL == cm)
        return;

    if (cm->getActorId() != layoutid)
        return;

    switch (cm->getMethod()) {
        case CrcfMessage::meGet:
            /* LAYOUT <layoutid> GET <attribute> */
            cms = getCrcfInfoMessage(cm->getAttribute());
            if (cms.isEmpty())
                emit statusMessage(tr("Unsupported CRCF attribute '%1' "
                            "detected.").arg(cm->getAttributeStr()));
            else
                sendGmCrcfMessage(sendto, replyto, cms);
            break;

        case CrcfMessage::meSet:
            /* LAYOUT <layoutid> SET <attribute> <att_value> */

            switch (cm->getAttribute()) {

                /* LAYOUT <layoutid> SET COLUMNS <att_value> */
                case CrcfMessage::atColumns:
                    if (visualMode == kvmEditLayout) {
                        int cc = cm->getAttValue();
                        if ((cc != cols) && (cc > 0) && (cc <= MAX_COLS)) {
                            setLayoutSize(cc, rows);
                            cms = getCrcfInfoMessage(CrcfMessage::atColumns);
                            if (!cms.isEmpty())
                                sendGmCrcfMessage(sendto, replyto, cms);
                            else
                                emit statusMessage(tr("Error assembling "
                                            "CRCF message for layout "
                                            "COLUMNS '%1'.").arg(cc));
                        }
                        else
                            emit statusMessage(tr("Unvalid COLUMNS "
                                        "value '%1' detected.").arg(cc));
                    }
                    else
                        emit statusMessage(tr("Layout data editing via "
                                    "CRCF messages is only allowed in "
                                    "layout edit mode."));
                    break;

                    /* LAYOUT <routeid> SET ROWS <att_value> */
                case CrcfMessage::atRows:
                    if (visualMode == kvmEditLayout) {
                        int cr = cm->getAttValue();
                        if (cr != rows && (cr > 0) && cr <= MAX_ROWS) {
                            setLayoutSize(cols, cr);
                            cms = getCrcfInfoMessage(CrcfMessage::atRows);
                            if (!cms.isEmpty())
                                sendGmCrcfMessage(sendto, replyto, cms);
                            else
                                emit statusMessage(tr("Error assembling "
                                            "CRCF message for layout "
                                            "ROWS '%1'.").arg(cr));
                        }
                        else
                            emit statusMessage(tr("Unvalid ROWS "
                                        "value '%1' detected.")
                                    .arg(cm->getAttValue()));
                    }
                    else
                        emit statusMessage(tr("Layout data editing via "
                                    "CRCF messages is only allowed in "
                                    "layout edit mode."));
                    break;

                    /* LAYOUT <layoutid> SET ID <att_value> */
                case CrcfMessage::atId:
                    if (visualMode == kvmEditLayout) {
                        unsigned int mid = cm->getAttValue();
                        if (layoutid != mid) {
                            setLayoutId(mid);
                            cms = getCrcfInfoMessage(CrcfMessage::atId);
                            if (!cms.isEmpty())
                                sendGmCrcfMessage(sendto, replyto, cms);
                            else
                                emit statusMessage(tr("Error assembling "
                                            "CRCF message for layout ID "
                                            "'%1'.").arg(mid));
                        }
                    }
                    else
                        emit statusMessage(tr("Layout data editing via "
                                    "CRCF messages is only allowed in "
                                    "layout edit mode."));
                    break;

                    /* LAYOUT <layoutid> SET NAME <att_value> */
                case CrcfMessage::atName:
                    if (visualMode == kvmEditLayout) {
                        QString ln = cm->getAttValueStr();
                        if (ln != layoutname) {
                            setLayoutName(ln);
                            cms = getCrcfInfoMessage(CrcfMessage::atName);
                            if (!cms.isEmpty())
                                sendGmCrcfMessage(sendto, replyto, cms);
                            else
                                emit statusMessage(tr("Error assembling "
                                            "CRCF message for layout NAME "
                                            "'%1'.").arg(ln));
                        }
                    }
                    else
                        emit statusMessage(tr("Layout data editing via "
                                    "CRCF messages is only allowed in "
                                    "layout edit mode."));
                    break;

                    /* LAYOUT <layoutid> SET TABLELIGHT <att_value> */
                case CrcfMessage::atTableLight:
                    {
                        unsigned int value = cm->getAttValue();
                        if (value > 1)
                            emit statusMessage(tr("Unvalid TABLELIGHT "
                                        "value '%1' detected.").arg(value));
                        else {
                            bool tl = (bool) value;
                            if (tl != tablelight) {
                                switchTableLight(tl);
                                cms = getCrcfInfoMessage(CrcfMessage::atTableLight);
                                if (!cms.isEmpty())
                                    sendGmCrcfMessage(sendto, replyto, cms);
                                else
                                    emit statusMessage(tr("Error assembling "
                                                "CRCF message for layout TABLELIGHT "
                                                "'%1'.").arg(tl));
                            }
                        }
                    }
                    break;

                default:
                    emit statusMessage(tr("Uneditable CRCF attribute '%1' "
                                "detected.").arg(cm->getAttributeStr()));
                    break;

            }// end SET switch
            break;

        default:
            emit statusMessage(tr("Unsupported CRCF method '%1' "
                        "detected.").arg(cm->getMethodStr()));
            break;
    }
}


/* send a SRCP GM CRCF message to server, method identical to router */
void GBSArea::sendGmCrcfMessage(unsigned int sendto,
        unsigned int replyto, const QString& cms)
{
    GmMessage sm = GmMessage(SrcpMessage::mtGmSet,
            sendto, replyto, "CRCF", cms);
    emit sendSrcpMessage(&sm);
}


/*assemble CRCF layout info message string*/
QString GBSArea::getCrcfInfoMessage(CrcfMessage::CrcfAttribute at) const
{
    unsigned int result = 0;

    switch (at) {
        case CrcfMessage::atId:
            result = layoutid;
            break;

        case CrcfMessage::atName:
            return CrcfMessage::message(CrcfMessage::acLayout, layoutid,
                    CrcfMessage::meInfo, at, layoutname);
            break;

        case CrcfMessage::atColumns:
            result = cols;
            break;

        case CrcfMessage::atRows:
            result = rows;
            break;

        case CrcfMessage::atTableLight:
            result = tablelight;
            break;

        default:
            return "";
            break;
    }

    return CrcfMessage::message(CrcfMessage::acLayout, layoutid,
            CrcfMessage::meInfo, at, result);
}

/*switch table light nn/off*/
void GBSArea::switchTableLight(bool ison)
{
    if (tablelight != ison) {
        tablelight = ison;
        for (int i = 0; i < elements.size(); i++) {
            TableLight* li = dynamic_cast<TableLight*>(elements[i]);
            if (li != NULL)
                li->setTableLight(ison);
        }
        if (tablelight)
            emit statusMessage(tr("Table light switched on"));
        else
            emit statusMessage(tr("Table light switched off"));
    }
}
